/*  mbus_init()
 *
 *  $Id: mbus.c,v 1.4 1998/02/21 02:01:01 kingdon Exp kingdon $
 */

#include <stdlib.h>
#include <rtems.h>
#include <bsp.h>
#include <cold.h>
#include <mbus.h>
#include <irq-map.h>

#define INTERRUPT_EVENT	RTEMS_EVENT_1

#define MEN	(1<<7)			/* M-Bus Enable */
#define MIEN	(1<<6)			/* M-Bus Interrupt Enable */
#define MSTA	(1<<5)			/* Master/Slave Mode Select */
#define MTX	(1<<4)			/* Transmit/Receive mode select */
#define TXAK	(1<<3)			/* Transmite Acknowledge Enable */
#define RSTA    (1<<2)                  /* perform repeated start */

#define MCF	(1<<7)			/* Data Transfering */
#define MAAS    (1<<6)                  /* addressed as slave */
#define MBB	(1<<5)			/* M-Bus Busy */
#define MAL	(1<<4)			/* Arbitration Lost */
#define MIF	(1<<1)			/* M-Bus Interrupt */
#define RXAK	(1<<0)			/* Received Acknowledge */

struct MBus_Info
{
      unsigned char* data;		/* R/W data */
      unsigned int   dlen;		/* length of data */
      unsigned int   dcnt;		/* bytes R/W */
      unsigned       is_read;
      unsigned       is_addr;		/* true only during addr xmt */
      rtems_id       lock_semaphore;	/* mutex lock */
      rtems_id       notify_pid;	/* process to notify */
      rtems_interval ticks_per_sec;
      unsigned int   transfer_result;   /* NACK?  OK? LAB? */
};

/*
 * Unless there is a way to get these codes back to 
 * the caller they are for diagnostic purposes only.
 *
 *                                            Tyson
 */

#define MBUS_RESULT_OK   0              /* No error */
#define MBUS_RESULT_LAB  1              /* Lost arbitration.  Noise? */
#define MBUS_RESULT_AAS  2              /* Addressed as slave */
#define MBUS_RESULT_NACK 3              /* byte not acked */
#define MBUS_RESULT_BUSY 4              /* bus busy */
#define MBUS_RESULT_ERR  5              /* some other error */

static struct MBus_Info MBus_Info[1];

#define MBUS_USE_ISR	1
#define MBUS_USE_POLL	0

#define MBUS_ISR_METHOD	MBUS_USE_ISR

static rtems_isr
MBus_isr( rtems_vector_number vector )
{
#if MBUS_ISR_METHOD == MBUS_USE_POLL
   volatile SIM5206*   sim  = (SIM5206*)MBAR_BASE;
   sim->mbus.mbsr = 0x00;
#else
   volatile SIM5206*   sim  = (SIM5206*)MBAR_BASE;
   struct MBus_Info*   info = &MBus_Info[0];
   unsigned            mbsr;
   unsigned            mbcr;

   /*
    * It may be advantagous to make better use of the 
    * local vars instead of re-reading the real registers.
    * This was originally the case but got changed during
    * hacking.
    *                                           Tyson
    */

   mbsr           = sim->mbus.mbsr;
   mbcr           = sim->mbus.mbcr;
   sim->mbus.mbsr &= ~MIF;		/* clear mbus interrupt flag */

#if 0
   RAW_FMT( "isr: mbsr=%x mbcr=%x dcnt=%d dlen=%d\n",
	    mbsr, mbcr, info->dcnt, info->dlen);
#endif

   /* check for spurious interrupt */

   if( !info->notify_pid ) {
     sim->mbus.mbcr &= ~MEN;
     sim->mbus.mbcr |= MEN;
     RAW_FMT( "mbus:INT?:sr=%x:cr=%x\n", mbsr, mbcr);
#if 0
     info->transfer_result = MBUS_RESULT_ERR; /* no one to rx this */
#endif
     goto end_of_packet;
   }

   /* check for LAB and AAS */

   if (!(mbcr & MSTA)) {  /* slave mode? */

     info->transfer_result = MBUS_RESULT_OK;

     if (mbsr & MAL) { /* mbus arbitration loss */
       sim->mbus.mbsr &= ~MAL;
       RAW_FMT( "mbus:LAB:mbsr=%x:mbcr=%x\n", mbsr, mbcr );
       info->transfer_result = MBUS_RESULT_LAB;
     }

     if (mbsr & MAAS) { /* have we been addressed? */
       sim->mbus.mbcr &= ~MEN;
       sim->mbus.mbcr |= MEN;
       RAW_FMT( "mbus:AAS:mbsr=%x:mbcr=%x\n", mbsr, mbcr );
       info->transfer_result = MBUS_RESULT_AAS;
     }

     if (info->transfer_result == MBUS_RESULT_OK) {
       RAW_FMT( "mbus:isr:strange slave mode:mbsr=%x:mbcr=%x\n", mbsr, mbcr );
       info->transfer_result = MBUS_RESULT_ERR;
     }

     if (info->dcnt) info->dcnt--;  /* keep dcnt accurate */
     goto end_of_packet;
   }

   /* Rx Mode */

   if( !(mbcr & MTX) ) {
     if (!info->is_read) { /* not a read message, why Rx? */
       sim->mbus.mbcr &= ~MEN;
       sim->mbus.mbcr |= MEN;
       RAW_FMT("mbus:isr:Tx message but Rx state:mbsr=%x:mbcr=%x\n",
	       mbsr, mbcr);
       info->transfer_result = MBUS_RESULT_ERR;
       goto end_of_packet;
     }
       
     if( info->dcnt > info->dlen ) {
       sim->mbus.mbcr &= ~MEN;
       sim->mbus.mbcr |= MEN;
       RAW_FMT( "mbus:isr:seems to have read too many bytes:mbsr=%x:mbcr=%x\n",
		mbsr, mbcr );
       info->transfer_result = MBUS_RESULT_ERR;
       goto end_of_packet;
     }
       
     if( info->dcnt+1 == info->dlen ) {
       sim->mbus.mbcr &= ~MSTA; /* clear MSTA to generate a stop */
     }

     if( info->dcnt+2 >= info->dlen ) {
       sim->mbus.mbcr |= TXAK; /* dont' ack last byte */
     }

     info->data[info->dcnt++] = sim->mbus.mbdr;
     if (info->dcnt < info->dlen) return;
     goto end_of_packet;
   }

   /* Tx mode */

   if(info->dcnt == info->dlen) {
     sim->mbus.mbcr &= ~MSTA; /* clear MSTA to generate a stop */
     info->transfer_result = MBUS_RESULT_OK;
     goto end_of_packet;
   }

   if(mbsr & RXAK) {      /* NACK */
     sim->mbus.mbcr &= ~MSTA; /* clear MSTA to generate a stop */
     info->transfer_result = MBUS_RESULT_NACK;
     if (info->dcnt) info->dcnt--;  /* keep dcnt accurate */
     goto end_of_packet;
   }

   if( info->is_read )  {
     unsigned junk;
     /* sent address: switch to recv mode... */
     info->is_addr = 0;
     sim->mbus.mbcr &= ~MTX;
     if( info->dlen == 1 ) {
       sim->mbus.mbcr |= TXAK; /* dont' ack only byte */
     }
     junk = sim->mbus.mbdr;
     return;
   }

   /* Tx a byte */
   if( info->is_addr ) info->is_addr = 0;
   sim->mbus.mbdr = info->data[info->dcnt++];
   return;

end_of_packet:

   sim->mbus.mbcr &= ~(MSTA|MTX|TXAK);
   if( info->notify_pid ) {
     rtems_event_send( info->notify_pid, INTERRUPT_EVENT );
     info->notify_pid = 0;
   }

#endif
}

static rtems_status_code
MBus_init( int use_isr, unsigned clock_speed )
{
   volatile SIM5206*   sim  = (SIM5206*)MBAR_BASE;
   register unsigned   oldsr;

   _CPU_ISR_Disable( oldsr );
   sim->mbus.mbcr = 0x00;		/* reset mbus */
   sim->imr      |= (1<<11);		/* disable mbus interrupts */
   _CPU_ISR_Enable( oldsr );

   switch( clock_speed/1000000 )
   {
      case 0:
      case 16:
      sim->mbus.mfdr = 0x0D;		/* /160 */
      break;
      case 25:
      sim->mbus.mfdr = 0x33;		/* /256 */
      break;
      case 33:
      sim->mbus.mfdr = 0x12;		/* /384 */
      break;
      default:
      return RTEMS_UNSATISFIED;
   }

   sim->mbus.mbsr = 0x00;		/* clear mif bit... */
   sim->mbus.madr = 0x00;		/* random slave address */

   /* initialize mbus interrupt control register */
   sim->icr11 = 0x80|(MBUS_IRQ_IL<<2)|MBUS_IRQ_IP;
   if( use_isr )
   {
      rtems_isr_entry   old_isr;
      rtems_interrupt_catch( MBus_isr, MBUS_IRQ_VEC, &old_isr );
      _CPU_ISR_Disable( oldsr );
      sim->imr       &= ~(1<<11);	/* enable mbus interrupts */
      sim->mbus.mbcr |= MIEN;		/* enable mbus interrupts */
      _CPU_ISR_Enable( oldsr );
   }

      
   sim->mbus.mbcr |= MEN;		/* enable mbus */
   return RTEMS_SUCCESSFUL;
}

static inline void
MBus_poll_until_eob( volatile mbusstruct* mbus )
{
   while( (mbus->mbsr & MIF) == 0 );
}

static inline int
MBus_busy( volatile mbusstruct* mbus )
{
   return (mbus->mbsr & MBB);
}

static inline void
MBus_stop( volatile mbusstruct* mbus )
{
   mbus->mbcr &= ~(MSTA|MTX);
}

static inline void
MBus_start( volatile mbusstruct* mbus, unsigned rw_addr )
{
   /* aquire the bus... */
   mbus->mbcr |= MSTA|MTX;

   /* send the address */
   mbus->mbdr = rw_addr;
}

/*
 * Special non-interrupt functions...
 * (these should only be used in emergency by non rtems code)
 */
static int
MBus_polled_init( unsigned rw_addr, unsigned clock_speed )
{
   volatile SIM5206*   sim = (SIM5206*)MBAR_BASE;
   unsigned stat;

   do
   {
      MBus_init( MBUS_USE_POLL, 0 );

      while( MBus_busy( &sim->mbus ) );
   
      MBus_start( &sim->mbus, rw_addr );
      
      MBus_poll_until_eob( &sim->mbus );
      stat = sim->mbus.mbsr;
      sim->mbus.mbsr = 0x00;	/* clear mif & mal */
   }
   while( stat & MAL );

   return (stat & RXAK) == 0;
}

int
MBus_polled_rcv( unsigned addr, unsigned char* data, int nbyte )
{
   volatile SIM5206*   sim = (SIM5206*)MBAR_BASE;
   int      nrcv = 0;
   unsigned stat;

   if( MBus_polled_init( (addr<<1) | 0x01, 0 ) )
   {
      /* switch to rcv mode... */
      sim->mbus.mbcr &= ~MTX;
      stat = sim->mbus.mbdr;	/* dummy read */

      while( nrcv < nbyte )
      {
	 MBus_poll_until_eob( &sim->mbus );
	 stat = sim->mbus.mbsr;
	 sim->mbus.mbsr = 0x00;	/* clear MIF & MAL */
	 if( stat & MAL )
	    break;
	 data[nrcv++] = sim->mbus.mbdr;

      }
   }

   MBus_stop( &sim->mbus );
   return nrcv;
}

int
MBus_polled_xmt( unsigned addr, unsigned char* data, int nbyte )
{
   volatile SIM5206*   sim = (SIM5206*)MBAR_BASE;
   int      nxmt = 0;
   unsigned stat;

   if( MBus_polled_init( (addr<<1) | 0x00, 0 ) )
   {
      while( nxmt < nbyte )
      {
	 sim->mbus.mbdr = data[nxmt++];
	 MBus_poll_until_eob( &sim->mbus );
	 stat = sim->mbus.mbsr;
	 sim->mbus.mbsr = 0x00;	/* clear MIF & MAL */
	 if( stat & RXAK )
	    break;
      }
   }

   MBus_stop( &sim->mbus );
   return nxmt;
}

rtems_device_driver
MBus_initialize(
   rtems_device_major_number major,
   rtems_device_minor_number minor,
   void *pargp
)
{
   rtems_status_code   status;
   struct MBus_Info*   info = &MBus_Info[0];

   status = rtems_semaphore_create( rtems_build_name( 'M', 'B', 'U', 'S' ),
				    1, RTEMS_DEFAULT_ATTRIBUTES, 0,
				    &info->lock_semaphore );
   if( status != RTEMS_SUCCESSFUL )
      rtems_panic( "Can't create MBus semaphore: %s\n",
		   rtems_status_text( status ) );

   (void)rtems_clock_get( RTEMS_CLOCK_GET_TICKS_PER_SECOND,
			  &info->ticks_per_sec );

#warning **** FIX ME ****
   rtems_io_register_name( "/dev/backlight", major, 0x2D );
   rtems_io_register_name( "/dev/contrast",  major, 0x2C );

   return MBus_init( MBUS_ISR_METHOD, 0 );
}

rtems_device_driver
MBus_noop(
  rtems_device_major_number major,
  rtems_device_minor_number minor,
  void *pargp
)
{
   return RTEMS_SUCCESSFUL;
}

#define MBUS_IO_RD	1
#define MBUS_IO_WR	0

static int
MBus_IO( unsigned addr,
	 unsigned char* data, signed int dlen,
	 unsigned is_read )
{
   volatile struct MBus_Info* info = &MBus_Info[0];
   volatile SIM5206*          sim = (SIM5206*)MBAR_BASE;
   int tryCounter;

   if( dlen < 1 )
      return 0;

   rtems_semaphore_obtain( info->lock_semaphore,
			   RTEMS_WAIT,
			   RTEMS_NO_TIMEOUT );

#if 0
   RAW_FMT( "%s(%x,%x,%d)...\n",
	    is_read ? "RdMBus" : "WrMBus",
	    addr, *data, dlen );
   _UART_flush();
#endif
   
#if MBUS_ISR_METHOD == MBUS_USE_ISR
#warning "MBus ISR not finished"
   /* perhapse this should be a critical section... */
   info->data    = data;
   info->dlen    = dlen;
   info->dcnt    = 0;
   info->is_addr = 1;
   info->is_read = is_read;
   info->transfer_result = MBUS_RESULT_OK;

   /* default to error */
   dlen = 0;
   tryCounter=0;
   info->transfer_result = MBUS_RESULT_OK;
   addr = (addr << 1) | (is_read ? 0x01 : 0x00);

   do {
     if ( !MBus_busy( &sim->mbus )) {
       rtems_status_code status;
       status = rtems_task_ident( RTEMS_SELF, 0, &info->notify_pid );
       if( status == RTEMS_SUCCESSFUL ) {
	 rtems_event_set pending;
	 MBus_start( &sim->mbus, addr );
	 status = rtems_event_receive(INTERRUPT_EVENT,
				      RTEMS_WAIT|RTEMS_EVENT_ANY,
				      info->ticks_per_sec,
				      &pending );

	 if( status != RTEMS_SUCCESSFUL ) {
	   sim->mbus.mbcr &= ~MEN;
	   sim->mbus.mbcr |= MEN;
	   RAW_FMT( "MBus_IO: status != RTEMS_SUCCESSFUL\n" );
	 }

	 dlen = info->dcnt;

	 if (((dlen == 0) || (info->transfer_result != MBUS_RESULT_OK)) &&
	     (info->transfer_result != MBUS_RESULT_NACK)) {
	   RAW_FMT( "mbus: dlen = %d  transfer_result = %d\n",
		    dlen, info->transfer_result);
	   _UART_flush();
	 }
       }
       else {
	 RAW_FMT( "MBus_IO: who am i? " );
       }
     }
     else {
       sim->mbus.mbcr &= ~MEN;
       sim->mbus.mbcr |= MEN;
       info->transfer_result = MBUS_RESULT_BUSY;
       /* can we try a stratigy of sending clock pulses here? */
       RAW_FMT( "mbus:BB: %d\n", tryCounter);
       _UART_flush();
     }
   } while ((tryCounter++<2) && (info->transfer_result != MBUS_RESULT_OK));

   if (((dlen == 0) || (info->transfer_result != MBUS_RESULT_OK)) &&
       (info->transfer_result != MBUS_RESULT_NACK)) {
     info->notify_pid = 0;
     RAW_FMT( "mbus: **** dlen = %d  transfer_result = %d ****\n",
	      dlen, info->transfer_result);
     _UART_flush();
   }

#else
#warning "MBus ISR not used!!!"
   if( is_read )
     dlen = MBus_polled_rcv( addr, data, dlen );
   else
     dlen = MBus_polled_xmt( addr, data, dlen );
#endif
#if 0
   RAW_FMT( "%s(%x/%x:%d)\n",
	    is_read ? "RdMBus" : "WrMBus",
	    addr, *data, dlen );
#endif

   rtems_semaphore_release( info->lock_semaphore );

   return dlen;
}

rtems_device_driver
MBus_write(
  rtems_device_major_number major,
  rtems_device_minor_number minor,
  void *arg
)
{
   rtems_libio_rw_args_t*     rw_args = arg;

   rw_args->bytes_moved = MBus_IO( (minor & 0xFF),
				   (unsigned char*)rw_args->buffer,
				   (signed int)rw_args->count,
				   MBUS_IO_WR );

   return (rw_args->bytes_moved > 0) ? RTEMS_SUCCESSFUL : RTEMS_UNSATISFIED;
}

rtems_device_driver
MBus_read(
  rtems_device_major_number major,
  rtems_device_minor_number minor,
  void *arg
)
{
   rtems_libio_rw_args_t*     rw_args = arg;

   rw_args->bytes_moved = MBus_IO( (minor & 0xFF),
				   (unsigned char*)rw_args->buffer,
				   (signed int)rw_args->count,
				   MBUS_IO_RD );

   return (rw_args->bytes_moved > 0) ? RTEMS_SUCCESSFUL : RTEMS_UNSATISFIED;
}

void
MBus_LCD_off( void )
{
   unsigned char off = 0x00;
   MBus_polled_xmt( 0x2C, &off, 1 ); /* turn off contrast */
   MBus_polled_xmt( 0x2D, &off, 1 ); /* turn off backlight */
}
